"""

"""
from sklearn import cluster as _cluster

from .._partial import _BigPartialFitMixin, _copy_partial_doc


@_copy_partial_doc
class PartialMiniBatchKMeans(_BigPartialFitMixin, _cluster.MiniBatchKMeans):
    pass
